# Copyright (C) 2008 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Export::Map;

use strict;
use GeoDB::Export;
use GeoDB::Utils;
use GeoDB::Export::Density;
use GD;
use QWizard;

my $have_qw = eval "require QWizard";

our @ISA = qw(GeoDB::Export GeoDB::Export::Density);

our $VERSION = "1.11";

our @GEOOPTIONS = 
  (
   { name => 'style',
     default => 'icons',
     values => [qw(icons dots 528-Circles)],
     type => 'menu',
     text => 'Waypoint Style:',
   },
   { name => 'icondir',
     doif => sub{qwparam('style') eq 'icons'},
     indent => 1,
     default => (($ENV{'PAR_TEMP'}) ? ($ENV{'PAR_TEMP'} . "/inc/icons/") : 
		 'dist/tomtom-icons/'),
     text => 'Icon subdirectory:',
     advanced => 1,
   },
   { name => 'iconscale',
     doif => sub{qwparam('style') eq 'icons'},
     indent => 1,
     values => [qw(auto 1 .5 .25)],
     default => 'auto',
     type => 'menu',
     text => 'Icon Scale:',
     help => 'Size to scale icon sizes in the map.  "auto" scales based on the number of waypoints to be shown.',
   },
   { name => 'maptiles',
     values => [qw(1 2 3 4)],
     default => 2,
     text => 'Number of Map Tiles to use:',
     helpdesc => '(a larger number means a larger map)',
   },
   { name => 'notext',
     type => 'checkbox',
     text => 'Do not add waypoint identifiers',
   },
   "",
   { name => 'dodensity',
     type => 'checkbox',
     text => 'Overlay a density plot',
     helpdesc => '(warning: this takes a while to calculate)',
   },
  );

my %sizemap=(
	     Micro => 'micro',
	     Small => 'small',
	     Regular => 'regular',
	     Large => 'large',
	     'Not Chosen' => 'nosize',
	    );

my %subtypemap=(
		'Traditional Cache' => 'regular',
		'Unknown Cache' => 'unknown',
		'Multi-cache' => 'multi',
		'Virtual Cache' => 'virtual',
		'Earthcache' => 'earth',
		'Letterbox Hybrid' => 'letterbox',
		'Earthcache' => 'earth',
	       );

sub map_it {
    my ($self, $data) = @_;
    return "waypoint.png" if ($data->{'type'} ne 'Geocache');
    if ($data->{'subtype'} eq 'Traditional Cache') {
	return "geocaches-" .
	  ($sizemap{$data->{'groundspeak_container'}} || 'nosize') .
	    "-" . $data->{'groundspeak_difficulty'} . ".png";
    }
    return "geocaches-" . 
	  ($subtypemap{$data->{'subtype'}} || 'unknown') .
	    "-" . $data->{'groundspeak_difficulty'} . ".png";
}

	
sub export_it {
     my ($self, $file, $set) = @_;

     GEODEBUG(1,"writing map to: $file\n");

     my $im = $self->make_export($set);

     open(O,">$file");
     print O $im->png;
     close(O);
}

sub make_export {
     my ($self, $set) = @_;

     my $count = $set->num();

     if ($count <= 0) {
	 GEOERR("No waypoints in the set: can't create a map"); 
     }

     my @bounds = $set->get_bounds();
     my ($im, $minlat, $minlon, $maxlat, $maxlon) =
       $self->make_map(@bounds,$self->{'options'}{'maptiles'}-1);
     my $red = $im->colorAllocate(255,0,0);
     my $black = $im->colorAllocate(0,0,0);
     my $redalpha = $im->colorAllocateAlpha(255,0,0,80);
     my ($xsize, $ysize) = $im->getBounds();

     $set->load_waypoint();

     my $scale = $self->{'options'}{'iconscale'};
     if ( $scale eq 'auto' ) {
	 GEODEBUG(5, "rescaling from auto to ...\n");
	 $scale = 1 if ($count > 0);
	 $scale = .5 if ($count > 100);
	 $scale = .25 if ($count > 200);
     }
     my $scaledsize = 20*$scale;
     my $scaledoffset = 10*$scale;
    GEODEBUG(5,"Creating map with scale='$scale', size=$scaledsize, scaledoffset=$scaledoffset\n");

     if ($self->{'options'}{'style'} eq 'icons') {

	 my $webdir = $self->{'options'}{'icondir'} ||
	   $self->get_config_dir() . "/icons/";
	 $webdir .= "/" if ($webdir !~ /\/$/);
	 my %icons;

	 $set->foreach(sub {
			   my $cache = shift;
			   my $data = $cache->{'data'};

			   my $mappedfile =
			     $self->map_it($data);
			   my $fullfile = $webdir . $mappedfile;
			   if (!exists($icons{$mappedfile})) {
			       $fullfile =
				 GeoDB::find_file_from_paths($mappedfile,
						      [$webdir, "/usr/share/geoqo/icons/"],
						      { parsubdir => 'icons' });
			       if (! -f $fullfile) {
				   GEOWARN("couldn't find file '$fullfile' \n");
			       } else {
				   $icons{$mappedfile} = 
				     GD::Image->newFromPng($fullfile);
			       }
			   }
			   if ($icons{$mappedfile}) {
			       if ($scale ne 1) {
				   $im->copyResized($icons{$mappedfile},

					     int($xsize*(($data->{'lon'}-$minlon)/($maxlon-$minlon)))-$scaledoffset,
					     int($ysize-$ysize*(($data->{'lat'}-$minlat)/($maxlat-$minlat)))-$scaledoffset,
					     0,0,$scaledsize,$scaledsize,21,21);
			       } else {
				   $im->copy($icons{$mappedfile},

					     int($xsize*(($data->{'lon'}-$minlon)/($maxlon-$minlon)))-10,
					     int($ysize-$ysize*(($data->{'lat'}-$minlat)/($maxlat-$minlat)))-10,
					     0,0,21,21);
			       }
			   }
			   if ($self->{'options'}{'notext'} ne '1') {
			       $im->string(gdTinyFont,
					   int($xsize*(($data->{'lon'}-$minlon)/($maxlon-$minlon)))-10,
					   int($ysize-$ysize*(($data->{'lat'}-$minlat)/($maxlat-$minlat)))-15,
					   $data->{'ident'},
					   $black);
			   }
		       });
     } elsif ($self->{'options'}{'style'} eq '528-Circles') {
	 my $radiusdist = 2 * .001428;  # right for lat, wrong for long
	 $set->foreach(sub {
			   my $cache = shift;
			   my $data = $cache->{'data'};

			   $im->filledEllipse(
					      int($xsize*(($data->{'lon'}-$minlon)/($maxlon-$minlon))),
					      int($ysize-$ysize*(($data->{'lat'}-$minlat)/($maxlat-$minlat))),
					      int($xsize*(($radiusdist)/($maxlon-$minlon)))+1,
					      int($ysize*(($radiusdist)/($maxlat-$minlat)))+1,
					  $redalpha);
		       });
     } else {
	 # regular dot plot
	 $set->foreach(sub {
			   my $cache = shift;
			   my $data = $cache->{'data'};

			   $im->rectangle(int($xsize*(($data->{'lon'}-$minlon)/($maxlon-$minlon))),
					  int($ysize-$ysize*(($data->{'lat'}-$minlat)/($maxlat-$minlat))),
					  int($xsize*(($data->{'lon'}-$minlon)/($maxlon-$minlon)))+1,
					  int($ysize-$ysize*(($data->{'lat'}-$minlat)/($maxlat-$minlat)))+1,
					  $red);
		       });
     }

     if ($self->{'options'}{'dodensity'}) {
	 my ($spreadresults, $maxcount,
	     $maxdenlat, $mindenlat, $maxdenlon, $mindenlon) = $self->get_density_map($set);
	 my ($binnumx, $binnumy) = ($#$spreadresults, $#{$spreadresults->[0]});
	 my $opaque = 100; # XXX: make setable

	 my $latdiff = $maxdenlat - $mindenlat;
	 my $londiff = $maxdenlon - $mindenlon;
	 my $dozeros = 0; # XXX: configurable

	 for (my $i = 0; $i <= $binnumx; $i++) {
	     for (my $j = 0; $j <= $binnumy; $j++) {

		 next if ($spreadresults->[$i][$j] == 0 || $dozeros);

		 my ($red, $green, $blue) = hsv_to_rgb((($maxcount == 0) ? 0 : .8*$spreadresults->[$i][$j] / $maxcount));

		 # cache the colors for repeated use of same?
		 my $color = $im->colorAllocateAlpha(hsv_to_rgb((($maxcount == 0) ? 0 : .8*$spreadresults->[$i][$j] / $maxcount)),$opaque);

		 my $lat1 = $mindenlat+$i/$binnumx*$latdiff;
		 my $lat2 = $mindenlat+($i+1)/$binnumx*$latdiff;
		 my $lon1 = $mindenlon+$j/$binnumx*$londiff;
		 my $lon2 = $mindenlon+($j+1)/$binnumx*$londiff;

		 $im->filledRectangle(int($xsize*(($lon1-$minlon)/($maxlon-$minlon))),
				int($ysize-$ysize*(($lat1-$minlat)/($maxlat-$minlat))),
				int($xsize*(($lon2-$minlon)/($maxlon-$minlon)))+1,
				int($ysize-$ysize*(($lat2-$minlat)/($maxlat-$minlat)))+1,
				$color);


	     }
	 }
     }

     return $im;
}

sub make_map {
    my ($self, $minlat, $minlon, $maxlat, $maxlon, $diff) = @_;
    my ($osmzoom, $x1, $y1, $x2, $y2) = getTileRange(@_);
    my $im = $self->get_map_image($osmzoom, $x1, $y1, $x2, $y2);
    return ($im, get_map_coords($osmzoom, $x1, $y1, $x2, $y2)) if (wantarray);
    return $im;
}

sub latlon_to_coord {
    
}

#sub plot_something {
#    my ($minlat, $minlon, $maxlat, $maxlon, $diff) = @_;
#
#    my ($homelon, $homelat) = (-121.699177874, 38.555848757);
#    my $im;
#    my $red = $im->colorAllocate(255,0,0);
#
#    my ($i, $x1, $y1, $x2, $y2) = getTileRange(@_);
#    maybemerge;
#
#    $im->rectangle(int($xsize*(($homelon-$minlon)/($maxlon-$minlon))),
#		   int($ysize-$ysize*(($homelat-$minlat)/($maxlat-$minlat))),
#		   int($xsize*(($homelon-$minlon)/($maxlon-$minlon)))+1,
#		   int($ysize-$ysize*(($homelat-$minlat)/($maxlat-$minlat)))+1,
#		   $red);
#
#    $im->rectangle(40,40,100,100,
#		   $red);
#
#
#    # from the map grid squares, calculate the min/max lat/lon that
#    # the map *actually* displays.
#
#    # XXX: we shouldn't need to do this odd sorting hack
#    my (@lats, @lons);
#    ($lats[0], $lons[0], $lats[1], $lons[1]) = Project($x, $y, $z);
#    ($lats[2], $lons[2], $lats[3], $lons[3]) = Project($x2, $y2, $z);
#
#    @lats = sort {$a <=> $b} @lats;
#    @lons = sort {$a <=> $b} @lons;
#
#    # the final resulting min/max lat/lon
#    my ($minlat, $minlon, $maxlat, $maxlon) =
#      ($lats[0], $lons[0], $lats[3], $lons[3]);
#
#    # Max sure they're really min/max properly
#    ($minlat, $maxlat) = swapit($minlat, $maxlat) if ($minlat > $maxlat);
#    ($minlon, $maxlon) = swapit($minlon, $maxlon) if ($minlon > $maxlon);
#
#    return $im;
#
#
#    open(O,">tmp.png");
#    print O $im->png;
#    close(O);
#}

# given OSM data get the min/max lat/lon ranges for the map
sub get_map_coords {
    my ($z, $x, $y, $x2, $y2) = @_;

    # this is a hack because picking the right ones directly failed
    # and I was too lazy to figure out why so am just sorting instead.
    my (@lats, @lons);
    ($lats[0], $lons[0], $lats[1], $lons[1]) = Project($x, $y, $z);
    ($lats[2], $lons[2], $lats[3], $lons[3]) = Project($x2, $y2, $z);

    @lats = sort {$a <=> $b} @lats;
    @lons = sort {$a <=> $b} @lons;

    return ($lats[0], $lons[0], $lats[3], $lons[3]);
}


#
# given a zoom level and a x/y tile range, grab the images from the
# openstreetmap server and combine them into a single tiled image.
# returns: GD::Image reference
#
sub get_map_image {
    my ($self, $z, $x, $y, $x2, $y2) = @_;

    $x2 ||= $x;
    $y2 ||= $y;

    # all OSM images are 256x256

    my $ysize = 256*($x2-$x+1);
    my $xsize = 256*($y2-$y+1);

    # 1 = create true color image not 256 mapped
    my $im = new GD::Image($xsize, $ysize, 1);
    my $white = $im->colorAllocate(255,255,255);

    my $mapdir = $self->get_config_dir() . "/mapdata/";
    mkdir ($mapdir) if (! -d $mapdir);

    GEODEBUG(5,"Map tile ranges from $x-$x2 and $y-$y2\n");
    foreach my $aa ($x..$x2) {
	foreach my $bb ($y..$y2) {

	    my $filename = "${z}_${aa}_${bb}.png";
	    my $overwrite = 0;
	    # XXX: set overwrite if older than 7 days?
	    grab_url("http://tile.openstreetmap.org/$z/$aa/$bb.png",
		     "$mapdir/$filename", $overwrite);

	    # implement disk cache cleanup routine

	    my $im2 = GD::Image->newFromPng("$mapdir/$filename",1);
 	    $im->copy($im2, ($aa-$x)*255+1, ($bb-$y)*255+1,
 		      1,1,256,256);
	}
    }
    return $im;
}

# utility to quickly swap(a,b) => (b,a)
sub swapit {
    return ($_[1], $_[0]);
}

# gets the range of a tile set
# when passed a min/max lat/lon and diff =(number of grid squares-1)
# it'll return the range and proper zoom setting for OpenStreetMap
# map server maps.

sub getTileRange {
    my ($self, $minlat, $minlon, $maxlat, $maxlon, $diff) = @_;
    # ok, this is sick and avoiding the math
    # need to redo this using real cals
    $diff ||= 0;

    # check that their ranges are worth beans
    ($minlat, $maxlat) = ($maxlat, $minlat) if ($minlat > $maxlat);
    ($minlon, $maxlon) = ($maxlon, $minlon) if ($minlon > $maxlon);
	
    # mapnik has max zoom of 18
    my $i;
    my ($x1, $y1);
    my ($x2, $y2);
    for ($i = 18; $i >= 0; $i--) {
	($x1, $y1) = getTileNumber($minlat, $minlon, $i);
	($x2, $y2) = getTileNumber($maxlat, $maxlon, $i);

	($x1, $x2) = ($x2, $x1) if ($x1 > $x2);
	($y1, $y2) = ($y2, $y1) if ($y1 > $y2);

	if (($y2-$y1) <= $diff && ($x2-$x1) <= $diff) {
	    last;
	}
    }
    return ($i, $x1, $y1, $x1+$diff, $y1+$diff);
}


#
# taken from the OSM wiki
#
use Math::Trig;
sub getTileNumber {
    my ($lat,$lon,$zoom) = @_;
    my $xtile = int( ($lon+180)/360 *2**$zoom ) ;
    my $ytile = int( (1 - log(tan($lat*pi/180) + sec($lat*pi/180))/pi)/2 *2**$zoom ) ;
    return(($xtile, $ytile));
}

sub Project {
    my ($X,$Y, $Zoom) = @_;
    my $Unit = 1 / (2 ** $Zoom);
    my $relY1 = $Y * $Unit;
    my $relY2 = $relY1 + $Unit;

    # note: $LimitY = ProjectF(degrees(atan(sinh(pi)))) = log(sinh(pi)+cosh(pi)) = pi
    # note: degrees(atan(sinh(pi))) = 85.051128..
    #my $LimitY = ProjectF(85.0511);

    # so stay simple and more accurate
    my $LimitY = pi;
    my $RangeY = 2 * $LimitY;
    $relY1 = $LimitY - $RangeY * $relY1;
    $relY2 = $LimitY - $RangeY * $relY2;
    my $Lat1 = ProjectMercToLat($relY1);
    my $Lat2 = ProjectMercToLat($relY2);
    $Unit = 360 / (2 ** $Zoom);
    my $Long1 = -180 + $X * $Unit;
    return(($Lat2, $Long1, $Lat1, $Long1 + $Unit)); # S,W,N,E
}

sub ProjectMercToLat($){
    my $MercY = shift();
    return( 180/pi* atan(sinh($MercY)));
}

sub ProjectF {
    my $Lat = shift;
    $Lat = deg2rad($Lat);
    my $Y = log(tan($Lat) + (1/cos($Lat)));
    return($Y);
}



1;

=head1 NAME

Map - Make a map

=head1 DESCRIPTION

START HELP

  Example:
  geoqo -s set:import:myPQ -e map:output.png

The map will show red dots whereever there is a cache (Yes virginia,
there will eventually be symbols and names).  The output (currently)
must be a PNG file.

END HELP

=cut

