# Copyright (C) 2007-2009 Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Set;

use GeoDB;
use strict;
use GeoDB::Utils;
use GeoDB::Waypoint;

our @ISA=qw(GeoDB);

our $VERSION = "1.11";

sub new {
    my $type = shift;
    my ($class) = ref($type) || $type;
    my %self = @_;
    bless(\%self, $class);
}

sub clone {
    my $self = shift;
    my $clone = {};
    %$clone = %$self;
    bless($clone, ref($self));
    $clone->{'caches'} = [];
    $self->foreach(sub { $_[1]->add($_[0]->clone()); }, $clone);
    return $clone;
}

sub add {
    my $self = shift;
    push @{$self->{'caches'}}, @_;
}

sub remove_by_ident {
    my ($self, $cacheid) = @_;
    @{$self->{'caches'}} =
      grep {$_->{'data'}{'ident'} ne $cacheid} @{$self->{'caches'}};
}

sub find_by_ident {
    my ($self, $cacheid) = @_;
    my @results =
      grep {$_->{'data'}{'ident'} eq $cacheid} @{$self->{'caches'}};
    return $results[0];
}

sub get {
    my ($self, $num) = @_;
    return $self->{'caches'}[$num];
}

sub add_by_ids {
    my $self = shift;
    foreach my $id (@_) {
	push @{$self->{'caches'}},
	  GeoDB::Waypoint->new(data => { ident => $id });
    }
}

sub sdisplay {
    my ($self, $dologs) = @_;
    my $out;
    foreach my $cache (@{$self->{'caches'}}) {
	$out .= $cache->sdisplay($dologs) . "\n";
    }
    return $out;
}

sub display {
    my $self = shift;
    print $self->sdisplay(@_);
}

sub foreach {
    my $self = shift;
    my $code = shift;
    my @args = @_;
    foreach my $cache (@{$self->{'caches'}}) {
	$code->($cache, @args);
    }
}

# do a bunch of common things easily
sub save {
    my $self = shift;
    GEODEBUG(1, "saving a set: $self->{'name'}\n");
    $main::dbh->begin_work();
    my ($next,$count) = (0,0);
    $self->setup_progress($self->num(), "Saving");
    $self->foreach(sub {
		       shift->save(@_);
		       if ($count++ % 1000 == 0) {
			   # commit every 100 saves
			   GEODEBUG(4, "committing $count\n");
			   $main::dbh->commit();
			   $main::dbh->begin_work();
		       }
		       $next = $self->set_progress($count)
			 if ($count >= $next);
		   });
    $main::dbh->commit();
    $self->stop_progress();
    $self->save_set($main::dbh) if ($self->{'name'});
}

sub remember {
    my $self = shift;

    GEODEBUG(1, "saving a set: $self->{'name'}\n");
    $self->foreach(sub { shift->save(@_)});
}

sub load {
    my $self = shift;
    GEODEBUG(3, "loading a set's data ($#{$self->{'caches'}} items)\n");
    $self->setup_progress($self->num(), "Loading Data");
    my ($next,$count) = (0,0);
    $self->foreach(sub { 
		       shift->load(@_);
		       $count++;
		       $next = $self->set_progress($count)
			 if ($count >= $next);
		   }, @_);
    $self->stop_progress();
}

sub load_parts {
    my $self = shift;
    GEODEBUG(3, "loading a set's data ($#{$self->{'caches'}} items)\n");
    $self->setup_progress($self->num(), "Loading Data");
    my ($next,$count) = (0,0);
    $self->foreach(sub {
		       shift->load_parts(@_);
		       $count++;
		       $next = $self->set_progress($count)
			 if ($count >= $next);
		   }, @_);
    $self->stop_progress();
}

sub load_waypoint {
    my $self = shift;
    GEODEBUG(3, "loading a set's data ($#{$self->{'caches'}} items)\n");
    $self->setup_progress($self->num(), "Loading Data");
    my ($next,$count) = (0,0);
    $self->foreach(sub { 
		       shift->load_waypoint(@_);
		       $count++;
		       $next = $self->set_progress($count)
			 if ($count >= $next);
		   }, @_);
    $self->stop_progress();
}

sub num {
    return 1+$#{$_[0]->{'caches'}};
}

sub merge {
    my ($self, $newset) = @_;
    my %caches;
    map { $caches{$_->{'data'}{'ident'}} = $_ } @{$self->{'caches'}};
    map { $caches{$_->{'data'}{'ident'}} = $_ } @{$newset->{'caches'}};
    $self->{'caches'} = [values(%caches)];
}

sub ssummary {
    my ($self, $options) = @_;
    GEODEBUG(3, "summarizing a set\n");
    my $out;
    if ($self->num() == 0) {
	return "Result set: empty\n";
    }
    $out .= sprintf("num GCID     Name\n") if (!exists($options->{'format'}));
    my $num = $self->{'startnum'} || 0;
    $self->foreach(sub {
		       my $self = shift;
		       my $countscalar = shift;
		       $$countscalar++;
		       $out .= $self->ssummary($$countscalar, @_);
		   }, \$num, $options);
    return $out;
}

sub summary { 
    my $self = shift;
    print $self->ssummary(@_);
}

sub set_name {
    my ($self, $name) = @_;
    $self->{'name'} = $name;
}

sub get_name {
    my ($self, $name) = @_;
    return $self->{'name'};
}


sub save_set {
    my ($self, $dbh, $name) = @_;
    $name ||= $self->{'name'};
    return if (!$name);

    if (!exists($self->{'appendset'})) {
	my $sth = $dbh->prepare_cached("delete from sets where name = ?");
	$sth->execute($name);
    }

    $dbh->begin_work();

    my $sth = $dbh->prepare_cached("insert into sets(name, ident)
                                                values(?,?)");
    my (@cacheids);
    foreach my $cache (@{$self->{'caches'}}) {
	push @cacheids, $cache->{'data'}{'ident'};
    }
    GEODEBUG(1,"saving " . ($#cacheids+1) . " cache ids as set '$name'\n");
    $sth->execute_array({ ArrayTupleStatus => \my @tuple_status },
			$name, \@cacheids);

    # save the timestamp information
    my $delinfo = $dbh->prepare_cached("delete from setinfo where name = ?");
    my $insinfo = $dbh->prepare_cached("insert into setinfo(name, modified)
                                                values(?,?)");
    $delinfo->execute($name);
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)
      = localtime(time());
    $insinfo->execute($name,
		      sprintf("%04d/%02d/%02d", $year+1900, $mon+1, $mday));

    $dbh->commit();
}

sub load_set {
    my ($self, $dbh, $name) = @_;
    $name ||= $self->{'name'};
    GEODEBUG(3,"loading $name\n");
    my $rows = select_flat1($dbh, "select ident from sets where name = ?",
			    $name);
    GEODEBUG(3,"loading into set\n");
    $self->add_by_ids(@$rows);
}

sub sortby {
    my ($self, $sortsub) = @_;
    @{$self->{'caches'}} = sort $sortsub @{$self->{'caches'}};
}

sub splice {
    my ($self, $offset, $limit) = @_;
    if ($limit) {
	splice(@{$self->{'caches'}}, $offset, $limit);
    } else {
	splice(@{$self->{'caches'}}, $offset);
    }
}

sub clear {
    my ($self) = @_;
    $self->{'caches'} = [];
    delete $self->{'startnum'};
}

sub remove_from_db {
    my $self = shift;
    GEODEBUG(3, "deleting waypoints from db ($#{$self->{'caches'}} items)\n");
    $self->foreach(sub {
		       my $cache = shift;
		       $cache->remove_from_db();
		   }, @_);
}

sub get_bounds {
    my $self = shift;
    my ($minlat, $minlon, $maxlat, $maxlon) = (181, 181, -181, -181);

    # ensure the data is loaded
    $self->load_waypoint();

    # loop through each waypoint and calculate the min/max
    $self->foreach(sub {
		       my $cache = shift;
		       my $data = $cache->{'data'};
		       $maxlat = $data->{'lat'} if (defined($data->{'lat'}) &&
						    $data->{'lat'} > $maxlat);
		       $maxlon = $data->{'lon'} if (defined($data->{'lon'}) &&
						    $data->{'lon'} > $maxlon);
		       $minlat = $data->{'lat'} if (defined($data->{'lat'}) &&
						    $data->{'lat'} < $minlat);
		       $minlon = $data->{'lon'} if (defined($data->{'lon'}) &&
						    $data->{'lon'} < $minlon);
		   });

    # save the results to the set
    ($self->{'minlat'}, $self->{'maxlat'},
     $self->{'minlat'}, $self->{'maxlat'}) =
       ($minlat, $minlon, $maxlat, $maxlon);

    return ($minlat, $minlon, $maxlat, $maxlon);
}

1;


